//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef INTERP_H
#define INTERP_H

#ifdef _MSC_VER
	#pragma once
#endif

#include <vd2/system/vdtypes.h>
#include "effect.h"

class IVDShaderLogOutput;

enum {
	kVDShaderInterpOp_end,
	kVDShaderInterpOp_mov,
	kVDShaderInterpOp_add,
	kVDShaderInterpOp_sub,
	kVDShaderInterpOp_mul,
	kVDShaderInterpOp_rsq,
	kVDShaderInterpOp_rcp,
	kVDShaderInterpOp_neg,
	kVDShaderInterpOp_texld,
	kVDShaderInterpOp_unpackb,
	kVDShaderInterpOp_unpackg,
	kVDShaderInterpOp_unpackr,
	kVDShaderInterpOp_unpacka
};

struct VDShaderInterpreterTransformedInsn {
	int op;
	int dst;
	int src1;
	int src2;
};

struct VDShaderScalarForm {
	VDShaderScalarForm();
	~VDShaderScalarForm();
	void LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out);

	typedef VDShaderInterpreterTransformedInsn TransformedInsn;
	typedef vdfastvector<TransformedInsn> Insns;
	Insns mInsns;

	vdfastvector<float> mConstants;
	int mSlotCount;		// total constants + temps
	int mInputBase;
	int mTempBase;

	struct InputMapping {
		ptrdiff_t mSrcOffset;
		ptrdiff_t mDstOffset;
	};

	int mOutputMappings[4];

	typedef vdfastvector<InputMapping> InputMappings;
	InputMappings mInputMappings;
};

struct VDShaderInterpreterSampler {
	const uint32 *mpSrc;
	ptrdiff_t	mPitch;
	float	mFWidth;
	float	mFHeight;
	float	mFWidth0;
	float	mFHeight0;
	uint32	mULimit;
	uint32	mVLimit;
	bool	mbClampU;
	bool	mbClampV;
	bool	mbBilinear;

	void	Init(const uint32 *src, ptrdiff_t pitch, uint32 w, uint32 h, bool clampU, bool clampV, bool bilinear);
	uint32	Sample(float u, float v) const;
};

class VDShaderInterpreter {
public:
	VDShaderInterpreter();
	~VDShaderInterpreter();

	void LoadSamplers(const VDShaderInterpreterSampler *samplers, int numSamplers);
	void LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out);
	void Execute(VDShaderConstant& result, const float *args);

protected:
	vdfastvector<float> mRegisterFile;

	const VDShaderInterpreterSampler *mpSamplers;

	VDShaderScalarForm	mScalarForm;
};

#endif
